"""Utils that use the MongoDB engine."""

from typing import Optional, Generator, Tuple

import pandas as pd
from ytpa_utils.val_utils import is_list_of_instances

from .mongodb_engine import MongoDBEngine


def get_mongodb_records_gen(database: str,
                            collection: str,
                            db_config: dict,
                            filter: Optional[dict] = None,
                            projection: Optional[dict] = None,
                            distinct: Optional[dict] = None) \
        -> Generator[pd.DataFrame, None, None]:
    """
    Prepare MongoDB feature generator with some options. See find_distinct_gen() and find_many_gen() for the format of
    the provided options.

    Provide (filter and/or projection) or distinct, but not both. Ex:
        - provide filter
        - provide projection
        - provide filter and projection
        - provide distinct

    If using 'distinct' input arg, filter is specified through the distinct dict:
        e.g. distinct = dict(group=<str>, filter=<dict>)

    Filter options:
        - equality: filter = dict(a='5')
        - set membership: filter = dict(b=[1, 2, 3])
        - MongoDB-formatted: filter = {'$gt': 50}
    """
    using_filt = filter is not None
    using_proj = projection is not None
    using_dist = distinct is not None
    using_filt_or_proj = using_filt or using_proj
    assert not (using_filt_or_proj and using_dist) # using one or the other, but not both

    engine = MongoDBEngine(db_config, database=database, collection=collection)

    if using_dist:
        assert 'group' in distinct
        return engine.find_distinct_gen(distinct['group'], filter=distinct.get('filter'))
    elif using_filt_or_proj:
        filter_for_req: dict = {}
        if filter:
            for key, val in filter.items():
                if isinstance(val, str): # equality
                    filter_for_req[key] = val
                elif is_list_of_instances(val, (str, int)): # set membership
                    filter_for_req[key] = {'$in': val}
                elif is_list_of_instances(val, list) and len(val) == 1 and len(val[0]) == 2: # a single range
                    filter_for_req[key] = {'$gte': val[0][0], '$lte': val[0][1]}
                elif isinstance(val, dict): # MongoDB-formatted
                    assert all(['$' in key_ for key_ in val])
                    filter_for_req[key] = val
                else:
                    raise NotImplementedError(f"Filter type not yet implemented: {key}: {val}.")

        return engine.find_many_gen(filter_for_req, projection=projection)
    else:
        raise NotImplementedError('You must provide at least one of the options (filter, projection, distinct).')


def load_all_recs_with_distinct(database: str,
                                collection: str,
                                db_config: dict,
                                group: str,
                                filter: Optional[dict] = None) \
        -> pd.DataFrame:
    """Load many distinct records (optional filter followed by distinct query)."""
    distinct_ = dict(group=group, filter=filter)  # filter is applied first
    df_gen = get_mongodb_records_gen(database, collection, db_config, distinct=distinct_)
    return pd.concat([df for df in df_gen], axis=0, ignore_index=True)



"""MongoDB Engine for CRUD and other ops"""

from typing import Dict, Union, Optional, Callable, List, Generator
import math

import pandas as pd

from pymongo import MongoClient
from pymongo.collection import Collection, ObjectId, Cursor
from pymongo.errors import BulkWriteError

from ytpa_utils.val_utils import is_list_of_instances

from .constants import MONGODB_FIND_MANY_MAX_COUNT



class MongoDBEngine():
    """Convenience class for interactions with a MongoDB database"""
    def __init__(self,
                 db_config: Dict[str, Union[str, int]],
                 database: Optional[str] = None,
                 collection: Optional[str] = None,
                 verbose: bool = False):
        self._db_config = db_config
        self._database = None
        self._collection = None
        self._verbose = verbose

        self._db_client = MongoClient(self._db_config['host'], self._db_config['port'])

        self.set_db_info(database=database, collection=collection)

        self._cursor = None # to ensure that client stays open in generators

    def __del__(self):
        self._db_client.close()

    def set_db_info(self,
                    database: Optional[str] = None,
                    collection: Optional[str] = None):
        """Set database and collection to be used in db op calls"""
        if database is not None:
            if database not in self.get_all_databases():
                if self._verbose:
                    print(f"MongoDBEngine.set_bg_info() -> Database {database} does not exist.")
            self._database = database
        if collection is not None:
            try:
                collections = [cn for cns in self.get_all_collections(database=database).values() for cn in cns]
                if collection not in collections:
                    if self._verbose:
                        print(f"MongoDBEngine.set_bg_info() -> Collection {collection} "
                              f"does not exist in database {database}.")
            except:
                if self._verbose:
                    print(f"MongoDBEngine.set_bg_info() -> Collection {collection} or database {database} "
                          f"does not exist.")
            self._collection = collection

    def get_db_info(self):
        """Get currently targeted database and collection"""
        return self._database, self._collection


    ### connection ###
    def get_db_config(self) -> Dict[str, str]:
        return self._db_config

    def _query_wrapper(self, func: Callable):
        """Wrapper for exception handling during MongoDB queries"""
        try:
            return func()
        except Exception as e:
            print(e)


    ## DB inspection ##
    def get_all_databases(self) -> List[str]:
        """Get all databases"""
        return self._db_client.list_database_names()

    def get_all_collections(self, database: Optional[str] = None) -> Dict[str, List[str]]:
        """Get all collections by database or just those for a specified database"""
        if database is not None:
            databases = [database]
        else:
            databases = self.get_all_databases()
        return {database: self._db_client[database].list_collection_names() for database in databases}

    def get_ids(self) -> List[str]:
        """Get all IDs for a collection"""
        # TODO: implement exception efficiently, e.g. write ids to temp collection, then read from there:
        #   https://stackoverflow.com/questions/27323107/mongodb-distinct-too-big-16mb-cap
        try:
            return [str(id) for id in self._get_collection().distinct('_id')]
        except:
            # exception if distinct is too large
            df_gen = self.find_many_gen(projection={'_id': 1})
            ids = [str(id_) for df_ in df_gen for id_ in df_['_id']]
            return ids

    ## DB operations ##
    def _get_collection(self) -> Collection:
        """Get collection object for queries"""
        assert self._database is not None
        assert self._collection is not None
        return self._db_client[self._database][self._collection]

    def insert_one(self, record: dict):
        """Insert one record"""
        def func():
            cn = self._get_collection()

            if ('_id' in record) and (cn.find_one({"_id": record['_id']}) is not None):
                raise Exception(f'MongoDBEngine: A record with id {record["_id"]} already exists in collection '
                                f'{self._collection} of database {self._database}.')

            res = cn.insert_one(record)

            if self._verbose:
                print(f'MongoDBEngine: Inserted {1} record with id {res.inserted_id} in collection {self._collection} '
                      f'of database {self._database}.')

        return self._query_wrapper(func)

    def insert_many(self, records: List[dict]):
        """Insert many records"""
        def func():
            try:
                cn = self._get_collection()
                cn.insert_many(records, ordered=False)
            except BulkWriteError as e:
                if self._verbose:
                    writeErrors = e.details['writeErrors']
                    print(f"Failed to write {len(writeErrors)} out of {len(records)} records.")
        return self._query_wrapper(func)

    def update_one(self,
                   filter: dict,
                   update: dict,
                   upsert: bool = False):
        """Update a single record"""
        def func():
            cn = self._get_collection()
            cn.update_one(filter, update, upsert=upsert)
        return self._query_wrapper(func)

    def update_many(self,
                    filter: dict,
                    update: List[dict],
                    upsert: bool = False,
                    max_pipeline_len: Optional[int] = 1000):
        """Update records"""
        def func():
            cn = self._get_collection()
            num_pipelines = math.ceil(len(update) / max_pipeline_len)
            for i in range(num_pipelines):
                update_i = update[i * max_pipeline_len: (i + 1) * max_pipeline_len]
                if self._verbose:
                    print(f'Updating {len(update_i)} records.')
                cn.update_many(filter, update_i, upsert=upsert)
        return self._query_wrapper(func)

    def find_one_by_id(self, id: str) -> Optional[dict]:
        """Find a single record"""
        def func():
            cn = self._get_collection()

            # try provided id as-is
            rec = cn.find_one({"_id": id})
            if rec is not None:
                return rec

            # try converting to ObjectId
            rec = cn.find_one({"_id": ObjectId(id)})
            if rec is not None:
                return rec

            # fail
            raise Exception(f'Could not find record with _id {id}.')

        return self._query_wrapper(func)

    def find_one(self,
                 filter: Optional[dict] = None,
                 projection: Optional[dict] = None) \
            -> dict:
        """Same as self.find_many_gen() but for a single record."""
        if filter is None:
            filter = {}

        def func():
            cn = self._get_collection()
            if projection is None:
                cursor = cn.find(filter, limit=1)
            else:
                cursor = cn.find(filter, projection, limit=1)
            return next(cursor, None)
        return self._query_wrapper(func)

    def find_many_by_ids(self,
                         ids: Optional[List[str]] = None,
                         limit: int = 0,
                         filter_other: Optional[dict] = None) \
            -> List[dict]:
        """Find many records"""
        def func():
            cn = self._get_collection()
            filter = {} if ids is None else {"_id": {"$in": ids}}
            if filter_other is not None:
                filter = {**filter, **filter_other}
            cursor = cn.find(filter, limit=limit)
            return [d for d in cursor]
        return self._query_wrapper(func)

    def find_many_gen(self,
                      filter: Optional[dict] = None,
                      projection: Optional[dict] = None) \
            -> Generator[pd.DataFrame, None, None]:
        """
        Generator of records given optional filter and projection arguments.

        Args:
        - 'filter' is a dict with options analogous to a WHERE clause in a SQL query.
            e.g. filter = dict(name={'$in': ['a', 'b']}).
        - 'projection' is a dict with fields to return, analogous to "SELECT item, status FROM ..."
          instead of "SELECT * FROM ..." in a SQL query:
            e.g. {"item": 1, "status": 1, "_id": 0} # this drops '_id' in the returned dict
        """
        if filter is None:
            filter = {}

        def func():
            cn = self._get_collection()
            if projection is None:
                self._cursor = cn.find(filter)
            else:
                self._cursor = cn.find(filter, projection)
            return self._df_generator()

        return self._query_wrapper(func)

    def find_with_group_gen(self,
                            group: dict,
                            filter: Optional[dict] = None) \
            -> Generator[pd.DataFrame, None, None]:
        """Find records using an aggregation pipeline"""
        def func():
            cn = self._get_collection()

            pipeline = []
            if filter is not None:
                pipeline += [filter]
            pipeline += [{"$group": group}]

            self._cursor = cn.aggregate(pipeline)
            return self._df_generator()

        return self._query_wrapper(func)

    def find_many(self,
                  filter: Optional[dict] = None,
                  projection: Optional[dict] = None) \
            -> pd.DataFrame:
        """Batch version of find_many_gen. Careful with size of returned dataframe."""
        return pd.concat([df for df in self.find_many_gen(filter=filter, projection=projection)], ignore_index=True)

    def find_distinct_gen(self,
                          field: str,
                          filter: Optional[dict] = None) \
            -> Generator[pd.DataFrame, None, None]:
        """
        Find all distinct values of a given field.
        Output is a generator of DataFrames that have one column whose label is the field input arg.

        e.g. filter = {'$match': {<key1>: <one_val>, <key2>: {'$in': <list_of_vals>}}}
        """
        assert filter is None or (len(filter) == 1 and '$match' in filter)
        def func():
            group = {"_id": "$" + field}
            for df in self.find_with_group_gen(group, filter=filter):
                yield df.rename(columns={'_id': field})
        return self._query_wrapper(func)

    def delete_many(self, ids: Union[List[str], dict]):
        """Delete records by id"""
        assert is_list_of_instances(ids, (str, ObjectId)) or ids == {}
        def func():
            cn = self._get_collection()
            filter = {"_id": {"$in": ids}} if isinstance(ids, list) else {}
            cn.delete_many(filter)
        return self._query_wrapper(func)

    def delete_all_records(self, confirm_delete: Optional[str] = None):
        """Delete all records in a collection"""
        if confirm_delete != 'yes':
            return
        def func():
            cn = self._get_collection()
            cn.delete_many({})
        return self._query_wrapper(func)

    def delete_all_records_in_database(self, database: str):
        """Delete all records in a specified database"""
        def func():
            for collection in self.get_all_collections(database=database)[database]:
                self.set_db_info(database=database, collection=collection)
                cn = self._get_collection()
                assert cn.database.name == database
                assert cn.name == collection
                cn.delete_many({})
        return self._query_wrapper(func)


    ## Helper methods ##
    def _df_generator(self) -> Generator[pd.DataFrame, None, None]:
        """
        Generator of DataFrames from records produced by iterating on a PyMongo cursor.

        Note: this method MUST be a member of the engine class. Otherwise an instance of the engine used as a generator
        may be garbage-collected before the generator can be used, causing a "cannot use client after closing" error.
        """
        while 1:
            recs: List[dict] = []
            for _ in range(MONGODB_FIND_MANY_MAX_COUNT):
                rec_ = next(self._cursor, None)
                if rec_ is None:
                    break
                recs.append(rec_)
            if recs:
                yield pd.DataFrame(recs)
            else:
                return





"""
Database utils for MySQL including basic CRUD operations and more
complex functionality using the engine.
"""

from typing import Dict, Optional, Callable, List, Union, Generator

import mysql.connector
import pandas as pd

from .constants import MYSQL_FETCH_MANY_MAX_COUNT




class MySQLEngine():
    """MySQL convenience class for CRUD and other operations on database records."""
    def __init__(self, db_config: Dict[str, str]):
        # members
        self._db_config = None

        # setup
        self.set_db_config(db_config)


    ### connection config ###
    def set_db_config(self, db_config: Dict[str, str]):
        self._db_config = db_config

    def get_db_config(self) -> Dict[str, str]:
        return self._db_config


    ### connection and exception handling ###
    def _get_connection(self, database: Optional[str] = None):
        """Establish connection with a MySQL database"""
        return mysql.connector.connect(
            host=self._db_config['host'],
            user=self._db_config['user'],
            password=self._db_config['password'],
            database=database
        )

    def _sql_query_wrapper(self,
                           func: Callable,
                           database: Optional[str] = None):
        """Wrapper for exception handling during MySQL queries"""
        try:
            with self._get_connection(database=database) as connection:
                with connection.cursor() as cursor:
                    return func(connection, cursor)
        except mysql.connector.Error as e:
            print(e)


    ### get database and table info ###
    def get_db_names(self) -> List[str]:
        """Get all names of existing databases"""
        def func(connection, cursor):
            cursor.execute("SHOW DATABASES")
            return [db_name[0] for db_name in cursor]
        return self._sql_query_wrapper(func)

    def describe_table(self,
                       database: str,
                       tablename: str) \
            -> List[tuple]:
        """Return schema of a table"""
        def func(connection, cursor):
            cursor.execute(f"DESCRIBE {tablename}")
            return cursor.fetchall()
        return self._sql_query_wrapper(func, database=database)



    ### pure SQL ###
    def execute_pure_sql(self,
                         database: str,
                         query: str):
        def func(connection, cursor):
            cursor.execute(query)
            connection.commit()

        self._sql_query_wrapper(func, database=database)



    ### operations on databases ###
    def create_db(self, db_name: str):
        """Create a database"""
        def func(connection, cursor):
            cursor.execute(f"CREATE DATABASE {db_name}")
            connection.commit()
        return self._sql_query_wrapper(func)

    def create_db_from_sql_file(self, filename: str):
        """Create a database from a .sql file with 'CREATE TABLE IF NOT EXISTS ...' statements"""
        assert '.sql' in filename
        def func(connection, cursor):
            with open(filename, 'r') as fd:
                sqlFile = fd.read()
            sqlCommands = sqlFile.split(';')
            for command in sqlCommands:
                if command.strip() != '':
                    cursor.execute(command)
            # with open(filename, 'r') as f:
            #     cursor.execute(f.read(), multi=True) # doesn't work...?
            connection.commit()
        return self._sql_query_wrapper(func)

    def drop_db(self, db_name: str):
        """Delete a database"""
        def func(connection, cursor):
            cursor.execute(f"DROP DATABASE {db_name}")
            connection.commit()
        return self._sql_query_wrapper(func)


    ### operations on tables ###
    def create_tables(self,
                      database: str,
                      queries: Union[str, List[str]]):
        """Create one or more tables in a specified database"""
        def func(connection, cursor):
            if isinstance(queries, str):
                cursor.execute(queries)
                connection.commit()
            else:
                for query in queries:
                    cursor.execute(query)
                    connection.commit()
        return self._sql_query_wrapper(func, database=database)

    def insert_records_to_table(self,
                                database: str,
                                query: str,
                                records: Optional[Union[str, List[tuple]]] = None):
        """Insert records into a table using a single query or a split query (instructions + raw_data)"""
        def func(connection, cursor):
            if records is None:
                cursor.execute(query)
            else:
                cursor.executemany(query, records)
            connection.commit()
        return self._sql_query_wrapper(func, database=database)

    def select_records(self,
                       database: str,
                       query: str,
                       mode: str = 'list',
                       tablename: Optional[str] = None,
                       cols: Optional[List[str]] = None,
                       as_generator: bool = False) \
            -> Union[Generator[pd.DataFrame, None, None], Generator[List[tuple], None, None], pd.DataFrame, List[tuple]]:
        """
        Retrieve records from a table.

        If mode == 'pandas', you must specify either the tablename (to infer all of that table's column names) or
        cols, which is the list of columns that the query will return. Either way, these column names will be used
        as the column names in the returned pandas dataframe.
        """
        assert mode in ['list', 'pandas']
        if mode == 'pandas':
            assert (tablename is None and cols is not None) or (tablename is not None and cols is None)
            if cols is None:
                cols = [e[0] for e in self.describe_table(database, tablename)]
        if mode == 'list':
            assert tablename is None

        if not as_generator:
            def func(connection, cursor):
                cursor.execute(query)
                records = cursor.fetchall() # if table empty, "1241 (21000): Operand should contain 1 column(s)"
                if mode == 'pandas':
                    return pd.DataFrame(records, columns=cols)
                return records

            return self._sql_query_wrapper(func, database=database)
        else:
            return self._select_records_gen(database, query, mode, cols=cols)

    def _select_records_gen(self,
                            database: str,
                            query: str,
                            mode: str = 'list',
                            cols: Optional[List[str]] = None):
        # sql_query_wrapper() doesn't work with yield...
        # Throws `mysql.connector.errors.ProgrammingError: 2055: Cursor is not connected`
        try:
            with self._get_connection(database=database) as connection:
                with connection.cursor() as cursor:
                    cursor.execute(query)
                    while (records := cursor.fetchmany(MYSQL_FETCH_MANY_MAX_COUNT)) is not None:
                        if not records:
                            return
                        if mode == 'pandas':
                            yield pd.DataFrame(records, columns=cols)
                        else:
                            yield records
        except mysql.connector.Error as e:
            print(e)

        # def func(connection, cursor):
        #     cursor.execute(query)
        #     while (records := cursor.fetchmany(100)) is not None:
        #         if mode == 'pandas':
        #             yield pd.DataFrame(records, columns=cols)
        #         else:
        #             yield records
        # return self._sql_query_wrapper(func, database=database)



    def select_records_with_join(self,
                                 database: str,
                                 tablename_primary: str,
                                 tablename_secondary: str,
                                 join_condition: str,
                                 cols_for_query: List[str],
                                 table_pseudoname_primary: Optional[str] = None,
                                 table_pseudoname_secondary: Optional[str] = None,
                                 where_clause: Optional[str] = None,
                                 limit: Optional[int] = None,
                                 cols_for_df: Optional[List[str]] = None,
                                 as_generator: bool = False) \
            -> Union[Generator[pd.DataFrame, None, None], pd.DataFrame]:
        """Select query on one table joined on second table"""
        if table_pseudoname_primary is None:
            table_pseudoname_primary = tablename_primary
        if table_pseudoname_secondary is None:
            table_pseudoname_secondary = tablename_secondary
        if cols_for_df is None:
            cols_for_df = cols_for_query
        query = (f"SELECT {', '.join(cols_for_query)} FROM {tablename_primary} as {table_pseudoname_primary} "
                 f"JOIN {tablename_secondary} as {table_pseudoname_secondary} ON {join_condition}")
        if where_clause is not None:
            query += f" WHERE {where_clause}"
        if limit is not None:
            query += f" LIMIT {limit}"

        return self.select_records(database, query, mode='pandas', cols=cols_for_df, as_generator=as_generator)



"""Utils implementing useful ops over MySQL engine."""

from typing import List, Optional, Dict, Tuple, Generator

import pandas as pd
from ytpa_utils.sql_utils import make_sql_query_where_one
from ytpa_utils.val_utils import is_subset

from .mysql_engine import MySQLEngine



def get_table_colnames(database: str,
                       tablename: str,
                       db_config: dict) \
        -> List[str]:
    """Get list of column names for a table in a database"""
    engine = MySQLEngine(db_config)
    table_info = engine.describe_table(database, tablename)
    colnames = [tup[0] for tup in table_info]
    return colnames


def get_table_primary_keys(database: str,
                           tablename: str,
                           db_config: dict) \
        -> List[str]:
    """Get list of primary-key column names for a table in a database"""
    engine = MySQLEngine(db_config)
    table_info = engine.describe_table(database, tablename)
    colnames = [tup[0] for tup in table_info if tup[3] == 'PRI']
    return colnames


def prep_keys_for_insert_or_update(database: str,
                                   tablename: str,
                                   data: dict,
                                   db_config: dict,
                                   keys: Optional[List[str]] = None) \
        -> List[str]:
    """Validation and prep of keys for insert and update ops"""
    # if keys not specified, get all column names for this table
    if keys is None:
        keys = get_table_colnames(database, tablename, db_config)

    # key list is non-empty
    assert len(keys) > 0

    # keys is a subset of dict keys
    is_subset(keys, data)

    # if multiple records, must have same number of records for all keys
    if isinstance(data[keys[0]], list):
        lens = [len(e) for e in data.values()]
        assert all([len_ == lens[0] for len_ in lens])

    return keys


def insert_records_from_dict(database: str,
                             tablename: str,
                             data: dict,
                             db_config: dict,
                             keys: Optional[List[str]] = None):
    """
    Insert all or a subset of the info from a dict to a database table. Subset is specified through 'keys' arg.

    Data dict could have individual entries or a list for each key. In the latter case, the number of entries must
    match for all keys.

    On duplicate key, do nothing.

    INSERT INTO table_name (column1, column2, column3, ...)
    VALUES (value1, value2, value3, ...)
    ON DUPLICATE KEY UPDATE <val_orig>=<val_orig>;
    """
    keys = prep_keys_for_insert_or_update(database, tablename, data, db_config, keys=keys)

    query = f"INSERT INTO {tablename} ({','.join(keys)}) VALUES (" + ','.join(['%s'] * len(keys)) + ")"
    query += f" ON DUPLICATE KEY UPDATE {keys[0]}={keys[0]}"

    if isinstance(data[keys[0]], list):
        records: List[tuple] = [tuple([data[key][i] for key in keys]) for i in range(len(data[keys[0]]))]
    else:
        records: List[tuple] = [tuple([data[key] for key in keys])]

    engine = MySQLEngine(db_config)
    engine.insert_records_to_table(database, query, records)



def update_records_from_dict(database: str,
                             tablename: str,
                             data: dict,
                             db_config: dict,
                             condition_keys: Optional[List[str]] = None,
                             keys: Optional[List[str]] = None,
                             another_condition: Optional[str] = None):
    """
    Same as insert_records_from_dict() but applying an update operation.

    UPDATE table_name
    SET column1 = value1, column2 = value2, ...
    WHERE condition;

    'condition_keys' are the columns that are used in the WHERE clause (which records to update)
    'keys' are the columns to be updated
    """
    keys = prep_keys_for_insert_or_update(database, tablename, data, db_config, keys=keys)

    if condition_keys is None:
        condition_keys = get_table_primary_keys(database, tablename, db_config)
    assert len(condition_keys) > 0
    is_subset(condition_keys, data)
    # assert len(set(condition_keys) - set(raw_data.keys())) == 0

    query = f"UPDATE {tablename} SET " + ', '.join([key + ' = %s' for key in keys])
    query += ' WHERE ' + ' AND '.join([key + ' = %s' for key in condition_keys])
    if another_condition is not None:
        query += ' AND ' + another_condition

    if isinstance(data[keys[0]], list):
        records: List[tuple] = [
            tuple(
                [data[key][i] for key in keys] + [data[key][i] for key in condition_keys]
            )
            for i in range(len(data[keys[0]]))
        ]
    else:
        records: List[tuple] = [tuple([data[key] for key in keys] + [data[key] for key in condition_keys])]

    engine = MySQLEngine(db_config)
    engine.insert_records_to_table(database, query, records)


def perform_join_mysql_query(db_config: dict,
                             database: str,
                             tablename_primary: str,
                             tablename_secondary: str,
                             table_pseudoname_primary: str,
                             table_pseudoname_secondary: str,
                             join_condition: str,
                             cols_all: Dict[str, List[str]],
                             filters: Optional[dict] = None,
                             limit: Optional[int] = None,
                             as_generator: bool = False) \
        -> Tuple[Generator[pd.DataFrame, None, None], MySQLEngine]:
    """
    Perform join query with specified options.

    Args:
        database: name of database to perform query on
        tablename_primary: first table in join
        tablename_secondary: second table in join
        table_pseudoname_primary: post-join name for first table
        table_pseudoname_secondary: post-join name for second table
        join_condition: join condition clause in query
        cols_all: dict with columns from the two tables to return (keys are table
                  pseudonames and keys are lists of corresponding column names)
        filters: dict with conditions for WHERE clause, e.g. filters = dict(username=['uname1', 'uname2']). See all
                 options for specifying sub-clauses. All sub-clauses are AND'd together.
        limit: max number of records to return
    """
    # column info for query
    cols_for_query = ([f'{table_pseudoname_primary}.{colname}' for colname in cols_all[table_pseudoname_primary]] +
                      [f'{table_pseudoname_secondary}.{colname}' for colname in cols_all[table_pseudoname_secondary]])
    cols_for_df = cols_all[table_pseudoname_primary] + cols_all[table_pseudoname_secondary]

    # where clause
    where_clause = None

    if filter is not None and len(filters) > 0:
        where_clauses = []
        for key, val in filters.items():
            # identify table that this condition applies to
            tablename_ = [tname for tname, colnames_ in cols_all.items() if key in colnames_]
            assert len(tablename_) == 1
            tablename_ = tablename_[0]

            # add where clause
            where_clauses.append(make_sql_query_where_one(tablename_, key, val))

        # join sub-clauses
        where_clause = ' AND '.join(where_clauses)

    # issue request
    engine = MySQLEngine(db_config)
    df = engine.select_records_with_join(
        database,
        tablename_primary,
        tablename_secondary,
        join_condition,
        cols_for_query,
        table_pseudoname_primary=table_pseudoname_primary,
        table_pseudoname_secondary=table_pseudoname_secondary,
        where_clause=where_clause,
        limit=limit,
        cols_for_df=cols_for_df,
        as_generator=as_generator
    )

    return df, engine
